/*  Autopackage GTK+ frontend
 *  Copyright (C) 2003  Hongli Lai <h.lai@chello.nl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>

#include <gtk/gtk.h>
#include "utils.h"
#include "protocol.h"
#include "receiver.h"
#include "processor.h"
#include "prefs.h"
#include "logview.h"
#include "eggstatusicon.h"

#include "prefix.h"

gchar *fifoName = NULL;			/* Filename of the FIFO */
GladeXML *xml = NULL;			/* The main Glade GUI */
const gchar *data_dir;
gint exitCode = 0;
#define W(x) glade_xml_get_widget (xml, x)
gboolean silentMode = FALSE;
gboolean finished = FALSE;
EggStatusIcon *statusIcon = NULL;


static gboolean
show (EggStatusIcon *icon, gpointer user_data)
{
	silentMode = FALSE;

	gtk_widget_show (W("install"));
	g_object_unref (statusIcon);
	statusIcon = NULL;

	if (!finished)
		pref_set ("silent", "false");
	return FALSE;
}


gboolean
hide ()
{
	char *s = lookup_file ("package-closed.png");

	silentMode = TRUE;
	pref_set ("silent", "true");

	statusIcon = egg_status_icon_new_from_file (s);
	egg_status_icon_set_tooltip (EGG_STATUS_ICON (statusIcon), _("Preparing packages..."), NULL);
	g_signal_connect (G_OBJECT (statusIcon), "activate", G_CALLBACK (show), NULL);

	gtk_widget_hide (W("install"));
	g_free(s);

	return FALSE;
}


gboolean
delete_install_window ()
{
	/* The user tried to close the window. Only close if the installation has finished. */
	return !finished;
}


/* Read data from the receiver thread and pass them to the data processor */
static gboolean
read_data (GIOChannel *input, GIOCondition cond, gpointer d)
{
	GString *buf = g_string_new ("");
	gchar *data;
	gboolean result;

	while (1)
	{
		gchar *line;

		g_io_channel_read_line (input, &line, NULL, NULL, NULL);
		if (!line || strcmp (line, "-\n") == 0)
		{
			g_free (line);
			break;
		}
		g_string_append (buf, line);
		g_free (line);
	}
	data = buf->str;
	g_string_free (buf, FALSE);

	/* #ifdef DEBUG
	g_print ("Main thread received:(\n%s)\n", data);
	#endif */

	result = process (data);

	g_free (data);
	readLock = FALSE;
	if (!result) exit (1);
	return TRUE;
}


static gboolean
do_pulse (GtkWidget *progress)
{
	gtk_progress_bar_pulse (GTK_PROGRESS_BAR (progress));
	return TRUE;
}


/* Start a pulsing progress bar */
void
pulse_start (gchar *widgetName)
{
	GObject *progress;
	guint id;

	progress = (GObject *) W(widgetName);
	id = (guint) GPOINTER_TO_INT (g_object_get_data (progress, "PULSEID"));
	if (id == 0)
	{
		id = gtk_timeout_add (20, (GtkFunction) do_pulse, progress);
		g_object_set_data (progress, "PULSEID", GINT_TO_POINTER ((gint) id));
	}
}


/* Stop a pulsing progress bar */
void
pulse_stop (gchar *widgetName)
{
	GObject *progress;
	guint id;

	progress = (GObject *) W(widgetName);
	id = (guint) GPOINTER_TO_INT (g_object_get_data (progress, "PULSEID"));
	if (id != 0)
	{
		g_source_remove (id);
		g_object_set_data (progress, "PULSEID", NULL);
	}
}


/* Is a progress bar pulsing? */
gboolean
pulsing (gchar *widgetName)
{
	GObject *progress;

	progress = (GObject *) W(widgetName);
	return ((guint) GPOINTER_TO_INT (g_object_get_data (progress, "PULSEID")) != 0);
}


/* Set the window title and tray icon tooltip */
void
set_title (gchar *format, ...)
{
	gchar *msg;
	va_list ap;

	va_start (ap, format);
	msg = g_strdup_vprintf (format, ap);
	va_end (ap);

	gtk_window_set_title (GTK_WINDOW (W("install")), msg);
	if (silentMode)
		egg_status_icon_set_tooltip (statusIcon, msg, NULL);
	g_free (msg);
}


static void
exit_func ()
{
	receiver_stop ();
}


static void
initialize ()
{
	GdkColor *white;
	GIOChannel *input;
	GError *error = NULL;

	/* Setup the GUI */
	xml = load_glade ("frontend.glade");
	white = &gtk_widget_get_style (W("titleBox"))->white;
	gtk_widget_modify_bg (W("titleBox"), GTK_STATE_NORMAL, white);
	//	gtk_window_set_default_size (GTK_WINDOW (W("install")), 600, 350);
	gtk_widget_show (W("logview"));

	/** DISABLED: for now, the gui should always show when invoked directly,
	if (pref_get_bool ("silent"))
		hide ();
		else*/

	
	gtk_widget_show (W("install")); 

	/* Listen for input from the FIFO */
	if (!g_thread_supported ()) g_thread_init (NULL);
	input = receiver_start (&error);
	if (!input)
	{
		show_error (NULL, _("System error: unable to start the receiver thread.\n%s"),
			error->message);
		g_error_free (error);
		exit (1);
	}
	g_atexit (exit_func);

	/* Watch for input */
	g_io_add_watch (input, G_IO_IN, read_data, NULL);

}


int
main (int argc, char *argv[])
{
	struct stat buf;

	/* Tell the launcher to exit as soon as GTK+ is initialized */
	gtk_init (&argc, &argv);
	if (g_getenv ("AUTOPACKAGE_GTKFE_XID") && g_getenv ("AUTOPACKAGE_GTKFE_ATOM"))
	{
		GdkEvent event;
		GdkNativeWindow win;

		win = (GdkNativeWindow) atoi (g_getenv ("AUTOPACKAGE_GTKFE_XID"));
		event.type = GDK_CLIENT_EVENT;
		event.client.type = GDK_CLIENT_EVENT;
		event.client.window = gdk_window_foreign_new (win);
		event.client.send_event = TRUE;
		event.client.message_type = gdk_atom_intern (g_getenv ("AUTOPACKAGE_GTKFE_ATOM"), FALSE);
		event.client.data_format = 8;
		strcpy (event.client.data.b, "DONE");
		gdk_event_send_client_message (&event, win);
		while (gtk_events_pending ()) gtk_main_iteration ();
	}


	/* Sanity-check arguments */
	if (argc != 2)
	{
		g_printerr (
			_("This program is used internally by Autopackage"
			" and should not be launched directly. "
			"Use autopackage-launcher-gtk instead.\n"));
		return 1;
	}

	if (stat (argv[1], &buf) != 0)
	{
		g_printerr ("ERROR: file %s not found.\n", argv[1]);
		return 1;
	} else if (!S_ISFIFO (buf.st_mode))
	{
		g_printerr ("ERROR: %s is not a FIFO!\n", argv[1]);
		return 1;
	}

	data_dir = DATADIR;
	TRACE("data_dir=%s\n", data_dir);

	fifoName = argv[1];
	initialize ();
	gtk_main ();

	return exitCode;
}
